import numpy as np
import pandas as pd
import scipy.stats as stats
from macrosynergy.learning.forecasting.model_systems import BaseRegressionSystem
from sklearn.linear_model import LinearRegression, Ridge

from macrosynergy.learning.forecasting import LADRegressor


class LinearRegressionSystem(BaseRegressionSystem):
    """
    Cross-sectional system of linear regression models for panel data.

    Parameters
    ----------
    fit_intercept : bool, default=True
        Whether to fit an intercept for each regression.
    positive : bool, default=False
        Whether to enforce positive coefficients for each regression.
    roll : int or str, default = "full"
        The lookback of the rolling window for the regression.
        This should be expressed in either integer units of the native
        dataset frequency, or as the string `roll = 'full'` to use the
        entire available history.
    min_xs_samples : int, default=2
        The minimum number of samples required in each cross-section training set
        for a regression model to be fitted on that cross-section. If `data_freq` is
        None or `unadjusted`, this parameter is specified in units of the underlying
        dataset frequency. Otherwise, this parameter should be expressed in units of the
        frequency specified in `data_freq`.
    data_freq : str, optional
        Training set data frequency for resampling. This is primarily to be used within
        the context of market beta estimation in the `BetaEstimator` class in
        `macrosynergy.learning`, allowing for cross-validation of the underlying dataset
        frequency for good beta estimation. Accepted strings are 'unadjusted' to use the
        native dataset frequency, 'W' for weekly, 'M' for monthly and 'Q' for quarterly.
        It is recommended to set this parameter to 'W', 'M' or 'Q' only when
        the native dataset frequency is greater.

    Notes
    -----
    Separate regression models are fit for each cross-section, but evaluation is performed
    over the panel. Consequently, the results of a hyperparameter search will choose
    a single set of hyperparameters for all cross-sections, but the model parameters
    themselves may differ across cross-sections.

    This estimator is primarily intended for use within the context of market beta
    estimation, but can be plausibly used for return forecasting or other downstream tasks.
    The `data_freq` parameter is particularly intended for cross-validating market beta
    estimation models, since choosing the underlying data frequency is of interest for
    this problem.
    """

    def __init__(
        self,
        fit_intercept=True,
        positive=False,
        roll="full",
        min_xs_samples=2,
        data_freq=None,
    ):
        # Call the parent class constructor
        super().__init__(roll=roll, data_freq=data_freq, min_xs_samples=min_xs_samples)

        # Additional checks
        self._check_init_params(
            fit_intercept,
            positive,
        )

        # Additional attributes
        self.fit_intercept = fit_intercept
        self.positive = positive

        # Create data structures to store model information for each cross-section
        self.coefs_ = {}
        self.intercepts_ = {}

    def create_model(self):
        """
        Instantiate a linear regression model.

        Returns
        -------
        LinearRegression
            A linear regression model with the specified hyperparameters.
        """
        return LinearRegression(
            fit_intercept=self.fit_intercept,
            positive=self.positive,
        )

    def store_model_info(self, section, model):
        """
        Store the coefficients and intercepts of a fitted linear regression model.

        Parameters
        ----------
        section : str
            The cross-section identifier.
        model : LinearRegression
            The fitted linear regression model.
        """
        self.coefs_[section] = model.coef_[0]
        self.intercepts_[section] = model.intercept_

    def _check_init_params(
        self,
        fit_intercept,
        positive,
    ):
        """
        Parameter checks for the LinearRegressionSystem constructor.

        Parameters
        ----------
        fit_intercept : bool
            Whether to fit an intercept for each regression.
        positive : bool
            Whether to enforce positive coefficients for each regression.
        """
        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")


class LADRegressionSystem(BaseRegressionSystem):
    """
    Cross-sectional system of LAD regression models.

    Parameters
    ----------
    fit_intercept : bool, default=True
        Whether to fit an intercept for each regression.
    positive : bool, default=False
        Whether to enforce positive coefficients for each regression.
    roll : int or str, default = "full"
        The lookback of the rolling window for the regression.
        This should be expressed in either integer units of the native
        dataset frequency, or as the string `roll = 'full'` to use the
        entire available history.
    min_xs_samples : int, default=2
        The minimum number of samples required in each cross-section training set
        for a regression model to be fitted on that cross-section. If `data_freq` is
        None or 'unadjusted', this parameter is specified in units of the underlying
        dataset frequency. Otherwise, this parameter should be expressed in unites of the
        frequency specified in `data_freq`.
    data_freq : str, optional
        Training set data frequency for resampling. This is primarily to be used within
        the context of market beta estimation in the `BetaEstimator` class in
        `macrosynergy.learning`, allowing for cross-validation of the underlying dataset
        frequency for good beta estimation. Accepted strings are 'unadjusted' to use the
        native dataset frequency, 'W' for weekly, 'M' for monthly and 'Q' for quarterly.
        It is recommended to set this parameter to 'W', 'M' or 'Q' only when
        the native dataset frequency is greater.

    Notes
    -----
    Separate regression models are fit for each cross-section, but evaluation is performed
    over the panel. Consequently, the results of a hyperparameter search will choose
    a single set of hyperparameters for all cross-sections, but the model parameters
    themselves may differ across cross-sections.

    This estimator is primarily intended for use within the context of market beta
    estimation, but can be plausibly used for return forecasting or other downstream tasks.
    The `data_freq` parameter is particularly intended for cross-validating market beta
    estimation models, since choosing the underlying data frequency is of interest for
    this problem.
    """

    def __init__(
        self,
        fit_intercept=True,
        positive=False,
        roll="full",
        min_xs_samples=2,
        data_freq=None,
    ):
        # Call the parent class constructor
        super().__init__(roll=roll, data_freq=data_freq, min_xs_samples=min_xs_samples)

        # Additional checks
        self._check_init_params(
            fit_intercept,
            positive,
        )

        # Additional attributes
        self.fit_intercept = fit_intercept
        self.positive = positive

        # Create data structures to store model information for each cross-section
        self.coefs_ = {}
        self.intercepts_ = {}

    def create_model(self):
        """
        Instantiate a LAD regression model.

        Returns
        -------
        LADRegressor
            A LAD regression model with the specified hyperparameters.
        """
        return LADRegressor(
            fit_intercept=self.fit_intercept,
            positive=self.positive,
        )

    def store_model_info(self, section, model):
        """
        Store the coefficients and intercepts of a fitted LAD regression model.

        Parameters
        ----------
        section : str
            The cross-section identifier.
        model : LADRegressor
            The fitted linear regression model.
        """
        self.coefs_[section] = model.coef_[0]
        self.intercepts_[section] = model.intercept_

    def _check_init_params(
        self,
        fit_intercept,
        positive,
    ):
        """
        Parameter checks for the LADRegressionSystem constructor.

        Parameters
        ----------
        fit_intercept : bool
            Whether to fit an intercept for each regression.
        positive : bool
            Whether to enforce positive coefficients for each regression.
        """
        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")


class RidgeRegressionSystem(BaseRegressionSystem):
    """
    Cross-sectional system of ridge regression models for panel data.

    Parameters
    ----------
    fit_intercept : bool, default=True
        Whether to fit an intercept for each regression.
    positive : bool, default=False
        Whether to enforce positive coefficients for each regression.
    alpha : float, default=1.0
        L2 regularization hyperparameter. Greater values specify stronger regularization.
    roll : int or str, default = "full"
        The lookback of the rolling window for the regression.
        This should be expressed in either integer units of the native
        dataset frequency, or as the string `roll = 'full'` to use the
        entire available history.
    tol : float, default=1e-4
        The tolerance for termination.
    solver : str, default='lsqr'
        Solver to use in the computational routines. Options are 'auto', 'svd', 'cholesky',
        'lsqr', 'sparse_cg', 'sag', 'saga' and 'lbfgs'.
    min_xs_samples : int, default=2
        The minimum number of samples required in each cross-section training set
        for a regression model to be fitted on that cross-section. If `data_freq` is
        None or 'unadjusted', this parameter is specified in units of the underlying
        dataset frequency. Otherwise, this parameter should be expressed in unites of the
        frequency specified in `data_freq`.
    data_freq : str, optional
        Training set data frequency for resampling. This is primarily to be used within
        the context of market beta estimation in the `BetaEstimator` class in
        `macrosynergy.learning`, allowing for cross-validation of the underlying dataset
        frequency for good beta estimation. Accepted strings are 'unadjusted' to use the
        native dataset frequency, 'W' for weekly, 'M' for monthly and 'Q' for quarterly.
        It is recommended to set this parameter to 'W', 'M' or 'Q' only when
        the native dataset frequency is greater.

    Notes
    -----
    Separate regression models are fit for each cross-section, but evaluation is performed
    over the panel. Consequently, the results of a hyperparameter search will choose
    a single set of hyperparameters for all cross-sections, but the model parameters
    themselves may differ across cross-sections.

    This estimator is primarily intended for use within the context of market beta
    estimation, but can be plausibly used for return forecasting or other downstream tasks.
    The `data_freq` parameter is particularly intended for cross-validating market beta
    estimation models, since choosing the underlying data frequency is of interest in
    quant analysis.
    """

    def __init__(
        self,
        fit_intercept=True,
        positive=False,
        alpha=1.0,
        tol=1e-4,
        solver="lsqr",
        roll="full",
        min_xs_samples=2,
        data_freq=None,
    ):
        # Call the parent class constructor
        super().__init__(roll=roll, data_freq=data_freq, min_xs_samples=min_xs_samples)

        # Checks
        self._check_init_params(
            alpha,
            fit_intercept,
            positive,
            tol,
            solver,
        )

        # Additional attributes
        self.alpha = alpha
        self.fit_intercept = fit_intercept
        self.positive = positive
        self.tol = tol
        self.solver = solver

        # Create data structures to store model information for each cross-section
        self.coefs_ = {}
        self.intercepts_ = {}

    def create_model(self):
        """
        Instantiate a ridge regression model.

        Returns
        -------
        Ridge
            A ridge regression model with the specified hyperparameters.
        """
        return Ridge(
            fit_intercept=self.fit_intercept,
            positive=self.positive,
            alpha=self.alpha,
            tol=self.tol,
            solver=self.solver,
        )

    def store_model_info(self, section, model):
        """
        Store the coefficients and intercepts of a fitted ridge regression model.

        Parameters
        ----------
        section : str
            The cross-section identifier.
        model : Ridge
            The fitted ridge regression model.
        """
        self.coefs_[section] = model.coef_[0]
        self.intercepts_[section] = model.intercept_

    def _check_init_params(
        self,
        alpha,
        fit_intercept,
        positive,
        tol,
        solver,
    ):
        """
        Parameter checks for the RidgeRegressionSystem constructor.

        Parameters
        ----------
        alpha : float
            L2 regularization hyperparameter. Greater values specify stronger
            regularization.
        fit_intercept : bool
            Whether to fit an intercept for each regression.
        positive : bool
            Whether to enforce positive coefficients for each regression.
        tol : float
            The tolerance for termination.
        solver : str
            Solver to use in the computational routines.
        """
        if not isinstance(alpha, (int, float)):
            raise TypeError("alpha must be either an integer or a float.")
        if alpha < 0:
            raise ValueError("alpha must be non-negative.")
        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")
        if not isinstance(tol, (int, float)):
            raise TypeError("tol must be either an integer or a float.")
        if tol <= 0:
            raise ValueError("tol must be a positive number.")
        if solver not in [
            "auto",
            "svd",
            "cholesky",
            "lsqr",
            "sparse_cg",
            "sag",
            "saga",
            "lbfgs",
        ]:
            raise ValueError(
                "solver must be one of 'auto', 'svd', 'cholesky', 'lsqr', 'sparse_cg', "
                "'sag', 'saga' or 'lbfgs'."
            )


class CorrelationVolatilitySystem(BaseRegressionSystem):
    """
    Cross-sectional system of moving average models to estimate correlation and volatility
    components of a macro beta separately over a panel of financial contracts.

    Parameters
    ----------
    correlation_lookback : int or str, default="full"
        The lookback of the rolling window for correlation estimation.
        If "full", the entire cross-sectional history is used. Otherwise, this parameter
        should be an integer specified in the native dataset frequency. If `data_freq` is
        not None or 'unadjusted', this parameter should be expressed in units of the
        frequency specified in `data_freq`.
    correlation_type : str, default='pearson'
        The type of correlation to be calculated. Accepted values are 'pearson', 'kendall'
        and 'spearman'.
    volatility_lookback : int or str, default="full"
        The lookback of the rolling window for volatility estimation.
        If "full", the entire cross-sectional history is used. Otherwise, this parameter
        should be an integer specified in the native dataset frequency. If `data_freq` is
        not None or 'unadjusted', this parameter should be expressed in units of the
        frequency specified in `data_freq`.
    volatility_window_type : str, default='rolling'
        The type of window to use for the volatility calculation. Accepted values are
        'rolling' and 'exponential'.
    min_xs_samples : int, default=2
        The minimum number of samples required in each cross-section training set
        for a regression model to be fitted on that cross-section. If `data_freq` is
        None or 'unadjusted', this parameter is specified in units of the underlying
        dataset frequency. Otherwise, this parameter should be expressed in unites of the
        frequency specified in `data_freq`.
    data_freq : str, optional
        Training set data frequency for resampling. This is primarily to be used within
        the context of market beta estimation in the `BetaEstimator` class in
        `macrosynergy.learning`, allowing for cross-validation of the underlying dataset
        frequency for good beta estimation. Accepted strings are 'unadjusted' to use the
        native dataset frequency, 'W' for weekly, 'M' for monthly and 'Q' for quarterly.
        It is recommended to set this parameter to 'W', 'M' or 'Q' only when
        the native dataset frequency is greater.

    Notes
    -----
    This class is specifically designed for market beta estimation based on the
    decomposition of the beta into correlation and volatility components in univariate
    analysis.

    Separate estimators are fit for each cross-section, but evaluation is performed
    over the panel. Consequently, the results of a hyperparameter search will choose
    a single set of hyperparameters for all cross-sections, but the model parameters
    themselves may differ across cross-sections.
    """

    def __init__(
        self,
        correlation_lookback="full",
        correlation_type="pearson",
        volatility_lookback="full",
        volatility_window_type="rolling",
        min_xs_samples=2,
        data_freq=None,
    ):
        # Call the parent class constructor
        super().__init__(
            roll="full", data_freq=data_freq, min_xs_samples=min_xs_samples
        )

        # Additional checks
        self._check_init_params(
            correlation_lookback,
            correlation_type,
            volatility_lookback,
            volatility_window_type,
        )

        # Additional attributes
        self.correlation_lookback = correlation_lookback
        self.correlation_type = correlation_type
        self.volatility_lookback = volatility_lookback
        self.volatility_window_type = volatility_window_type

        # Create data structures to store the estimated betas for each cross-section
        self.coefs_ = {}

    def _fit_cross_section(self, section, X_section, y_section):
        """
        Fit correlation and volatility estimators on a single cross-section.

        Parameters
        ----------
        section : str
            The identifier of the cross-section.
        X_section : pd.DataFrame
            Input feature matrix for the cross-section.
        y_section : pd.Series
            Target variable for the cross-section.
        """
        # First estimate local correlation between the benchmark and contract returns
        if self.correlation_lookback == "full":
            if self.correlation_type == "pearson":
                corr = np.corrcoef(X_section.values[:, 0], y_section.values)[0, 1]
            elif self.correlation_type == "spearman":
                X_section_ranked = np.argsort(np.argsort(X_section.values[:, 0]))
                y_section_ranked = np.argsort(np.argsort(y_section.values))
                corr = np.corrcoef(X_section_ranked, y_section_ranked)[0, 1]
            elif self.correlation_type == "kendall":
                corr = stats.kendalltau(X_section.values[:, 0], y_section.values)[0]
        else:
            X_section_corr = X_section.values[-self.correlation_lookback :, 0]
            y_section_corr = y_section.values[-self.correlation_lookback :]
            if self.correlation_type == "pearson":
                corr = np.corrcoef(X_section_corr, y_section_corr)[0, 1]
            elif self.correlation_type == "spearman":
                X_section_ranked = np.argsort(np.argsort(X_section_corr))
                y_section_ranked = np.argsort(np.argsort(y_section_corr))
                corr = np.corrcoef(X_section_ranked, y_section_ranked)[0, 1]
            elif self.correlation_type == "kendall":
                corr = stats.kendalltau(X_section_corr, y_section_corr)[0]

        # Now estimate local standard deviations of both the benchmark and contract returns
        if self.volatility_lookback == "full":
            X_section_std = X_section.values[:, 0].std(ddof=1)
            y_section_std = y_section.values.std(ddof=1)
        else:
            if self.volatility_window_type == "rolling":
                X_section_std = X_section.values[-self.volatility_lookback :, 0].std(
                    ddof=1
                )
                y_section_std = y_section.values[-self.volatility_lookback :].std(
                    ddof=1
                )
            elif self.volatility_window_type == "exponential":
                alpha = 2 / (self.volatility_lookback + 1)
                weights = np.array(
                    [(1 - alpha) ** i for i in range(len(X_section))][::-1]
                )
                X_section_std = np.sqrt(
                    np.cov(X_section.iloc[:, 0].values.flatten(), aweights=weights)
                )
                y_section_std = np.sqrt(np.cov(y_section.values, aweights=weights))

        # Get beta estimate and store it
        beta = corr * (y_section_std / X_section_std)
        self.store_model_info(section, beta)

    def predict(
        self,
        X,
    ):
        """
        Make naive zero predictions over a panel dataset.

        Parameters
        ----------
        X : pd.DataFrame
            Input feature matrix.

        Returns
        -------
        predictions : pd.Series
            Pandas series of zero predictions, multi-indexed by cross-section and date.

        Notes
        -----
        This method outputs zero predictions for all cross-sections and dates, since the
        CorrelationVolatilitySystem is solely used for beta estimation and no forecasting
        is performed.
        """
        # Checks
        if not isinstance(X, pd.DataFrame):
            raise TypeError("The X argument must be a pandas DataFrame.")
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("X must be multi-indexed.")
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")
        if not np.all(X.columns == self.feature_names_in_):
            raise ValueError(
                "The input feature matrix must have the same columns as the",
                "training feature matrix.",
            )
        if len(X.columns) != self.n_features_in_:
            raise ValueError(
                "The input feature matrix must have the same number of",
                "columns as the training feature matrix.",
            )
        if X.isnull().values.any():
            raise ValueError(
                "The input feature matrix must not contain any missing values."
            )
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise ValueError(
                "All columns in the input feature matrix for CorrelationVolatilitySystem",
                " must be numeric.",
            )

        predictions = pd.Series(index=X.index, data=0)

        return predictions

    def store_model_info(self, section, beta):
        """
        Store the betas induced by the correlation and volatility estimators.

        Parameters
        ----------
        section : str
            The cross-section identifier.
        beta : numbers.Number
            The beta estimate for the associated cross-section.
        """
        self.coefs_[section] = beta

    def create_model(self):
        """
        Redundant method for the CorrelationVolatilitySystem class.
        """
        raise NotImplementedError("This method is not implemented for this class.")

    def _check_xs_dates(self, min_xs_samples, num_dates):
        """
        Cross-sectional availability check.

        Parameters
        ----------
        min_xs_samples : int
            The minimum number of samples required in each cross-section training set for
            correlation and volatility estimation.
        num_dates : int
            The number of unique dates in the cross-section.

        Returns
        -------
        bool
            True if the number of samples is sufficient, False otherwise
        """
        if num_dates < min_xs_samples:
            return False
        # If the correlation lookback is greater than the number of available dates, skip
        # to the next cross-section
        if (
            self.correlation_lookback != "full"
            and num_dates <= self.correlation_lookback
        ):
            return False
        # If the volatility lookback is greater than the number of available dates, skip
        # to the next cross-section
        if self.volatility_lookback != "full" and num_dates <= self.volatility_lookback:
            return False

        return True

    def _check_init_params(
        self,
        correlation_lookback,
        correlation_type,
        volatility_lookback,
        volatility_window_type,
    ):
        """
        Parameter checks for the CorrelationVolatilitySystem constructor.

        Parameters
        ----------
        correlation_lookback : int or str
            The lookback of the rolling window for correlation estimation.
        correlation_type : str
            The type of correlation to be calculated.
        volatility_lookback : int or str
            The lookback of the rolling window for volatility estimation.
        volatility_window_type : str
            The type of window to use for the volatility calculation.
        """
        if not isinstance(correlation_lookback, (int, str)):
            raise TypeError("correlation_lookback must be an integer or a string.")
        if isinstance(correlation_lookback, int) and correlation_lookback <= 0:
            raise ValueError("correlation_lookback must be a positive integer.")
        if isinstance(correlation_lookback, str) and correlation_lookback != "full":
            raise ValueError(
                "correlation_lookback must be 'full' or a positive integer."
            )
        if not isinstance(correlation_type, str):
            raise TypeError("correlation_type must be a string.")
        if correlation_type not in ["pearson", "kendall", "spearman"]:
            raise ValueError(
                "correlation_type must be one of 'pearson', 'kendall' or 'spearman'."
            )
        if not isinstance(volatility_lookback, (int, str)):
            raise TypeError("volatility_lookback must be an integer or a string.")
        if isinstance(volatility_lookback, int) and volatility_lookback <= 0:
            raise ValueError("volatility_lookback must be a positive integer.")
        if isinstance(volatility_lookback, str) and volatility_lookback != "full":
            raise ValueError(
                "volatility_lookback must be 'full' or a positive integer."
            )
        if not isinstance(volatility_window_type, str):
            raise TypeError("volatility_window_type must be a string.")
        if volatility_window_type not in ["rolling", "exponential"]:
            raise ValueError(
                "volatility_window_type must be one of 'rolling' or 'exponential'."
            )


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from sklearn.linear_model import LinearRegression

    import macrosynergy.management as msm
    from macrosynergy.management import make_qdf

    np.random.seed(1)

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "BENCH_XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2013-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2010-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2010-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["BENCH_XR"] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2010-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2010-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2010-01-01", "2020-12-31", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {"GBP": ["2009-01-01", "2012-06-30"], "CAD": ["2018-01-01", "2100-01-01"]}
    dfd = msm.reduce_df(df=dfd, cids=cids, xcats=xcats, blacklist=black)

    dfd = dfd.pivot(index=["cid", "real_date"], columns="xcat", values="value")

    # Demonstration of CorrelationVolatilitySystem usage

    # X2 = pd.DataFrame(dfd["BENCH_XR"])
    # y2 = dfd["XR"]
    # cv = CorrelationVolatilitySystem().fit(X2, y2)
    # print(cv.coefs_)

    # # # Demonstration of LinearRegressionSystem usage
    # X1 = dfd.drop(columns=["XR", "BENCH_XR"])
    # y1 = dfd["XR"]
    # lr = LinearRegressionSystem(roll=5).fit(X1, y1)
    # print(lr.coefs_)
    
    X1 = dfd.drop(columns=["XR", "BENCH_XR"])
    y1 = dfd["XR"]
    lr = LADRegressionSystem().fit(X1, y1)
    print(lr.coefs_)



"""
Function for visualising a facet grid of time line charts of one or more categories.

.. code-block:: python

    import macrosynergy.visuals as msv
    ...
    msv.view.timelines(df, xcats=["FXXR","EQXR", "IR"], cids=["USD", "EUR", "GBP"] )
    ...
    msv.FacetPlot(df).lineplot(cid_grid=True)

"""

from typing import Dict, List, Optional, Tuple, Union

import pandas as pd

from macrosynergy.management.utils import reduce_df
from macrosynergy.visuals import FacetPlot, LinePlot
from numbers import Number

IDX_COLS: List[str] = ["cid", "xcat", "real_date"]


def timelines(
    df: pd.DataFrame,
    xcats: Optional[List[str]] = None,
    cids: Optional[List[str]] = None,
    intersect: bool = False,
    val: str = "value",
    cumsum: bool = False,
    start: str = None,
    end: Optional[str] = None,
    ncol: int = 3,
    square_grid: bool = False,
    legend_ncol: int = 1,
    same_y: bool = True,
    all_xticks: bool = False,  # ~(same_x) basically
    xcat_grid: bool = False,
    xcat_labels: Union[Optional[List[str]], Dict] = None,
    cid_labels: Union[Optional[List[str]], Dict] = None,
    sort_cid_labels: bool = False,
    single_chart: bool = False,
    label_adj: float = 0.05,
    title: Optional[str] = None,
    title_adj: float = 0.95,
    title_xadj: float = 0.5,
    title_fontsize: int = 22,
    cs_mean: bool = False,
    size: Tuple[Number, Number] = (12, 7),
    aspect: Number = 1.7,
    height: Number = 3.0,
    legend_fontsize: int = 12,
    blacklist: Dict = None,
    ax_hline: Union[float, Dict] = None,
):
    """
    Displays a facet grid of time line charts of one or more categories.

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized DataFrame with the necessary columns: 'cid', 'xcat', 'real_date'
        and at least one column with values of interest.
    xcats : List[str]
        extended categories to plot. Default is all in DataFrame.
    cids : List[str]
        cross sections to plot. Default is all in DataFrame. If this contains only one
        cross section a single line chart is created.
    intersect : bool
        if True only retains cids that are available for all xcats. Default is False.
    val : str
        name of column that contains the values of interest. Default is 'value'.
    cumsum : bool
        plot cumulative sum of the values over time. Default is False.
    start : str
        earliest date in ISO format. Default is earliest date available.
    end : str
        latest date in ISO format. Default is latest date available.
    ncol : int
        number of columns in facet grid. Default is 3.
    legend_ncol : int
        number of columns in legend. Default is 1.
    same_y : bool
        if True (default) all plots in facet grid share same y axis.
    all_xticks : bool
        if True x-axis tick labels are added to all plots in grid. Default is False, i.e
        only the lowest row displays the labels.
    xcat_grid : bool
        if True, shows a facet grid of line charts for each xcat for given cross
        sections. Default is False.
    xcat_labels : Union[Optional[List[str]], Dict]
        labels to be used for xcats. If not defined, the labels will be identical to
        extended categories.
    cid_labels : Union[Optional[List[str]], Dict]
        labels to be used for cids. If not defined, the labels will be identical to
        cross-sections.
    sort_cid_labels : bool
        if True, sorts the cross-sectional labels in the grid alphabetically. Otherwise,
        the order of `cids` is preserved. Default is False.
    single_chart : bool
        if True, all lines are plotted in a single chart.
    title : str
        chart heading. Default is no title.
    title_adj : float
        parameter that sets top of figure to accommodate title. Default is 0.95.
    title_xadj : float
        parameter that sets x position of title. Default is 0.5.
    title_fontsize : int
        font size of title. Default is 16.
    label_adj : float
        parameter that sets bottom of figure to fit the label. Default is 0.05.
    cs_mean : bool
        if True this adds a line of cross-sectional averages to the line charts. This is
        only allowed for function calls with a single category. Default is False.
    size : Tuple[Number, Number]
        two-element tuple setting width/height of single cross section plot. Default is
        (12, 7). This is irrelevant for facet grid.
    aspect : Number
        width-height ratio for plots in facet. Default is 1.7.
    height : Number
        height of plots in facet. Default is 3.
    legend_fontsize : int
        font size of legend. Default is 12.
    blacklist : dict
        cross-sections with date ranges that should be excluded from the dataframe.
    """

    if not isinstance(df, pd.DataFrame):
        raise TypeError("`df` must be a pandas DataFrame.")

    if len(df.columns) < 4:
        df = df.copy().reset_index()

    if val not in df.columns:
        if len(df.columns) == len(IDX_COLS) + 1:
            val: str = list(set(df.columns) - set(IDX_COLS))[0]
            if not pd.api.types.is_numeric_dtype(df[val]):
                raise ValueError(
                    f"Column '{val}' (passed as `metric`) is not numeric, and there are "
                    f"no other numeric columns in the DataFrame."
                )
        else:
            raise ValueError(
                f"Column '{val}' (passed as `metric`) does not exist, and there are "
                "none/many other numeric columns in the DataFrame."
            )

    if start is None:
        start: str = pd.Timestamp(df["real_date"].min()).strftime("%Y-%m-%d")

    if end is None:
        end: str = pd.Timestamp(df["real_date"].max()).strftime("%Y-%m-%d")

    if isinstance(xcats, str):
        xcats: List[str] = [xcats]
    if isinstance(cids, str):
        cids: List[str] = [cids]

    for varx, namex in zip([single_chart, xcat_grid], ["single_chart", "xcat_grid"]):
        if not isinstance(varx, bool):
            raise TypeError(f"`{namex}` must be a boolean.")

    if xcat_grid and single_chart:
        raise ValueError(
            "`xcat_grid` and `single_chart` cannot be True simultaneously."
        )

    if cs_mean and xcat_grid:
        raise ValueError("`cs_mean` requires `xcat_grid` to be False.")

    if blacklist:
        if not isinstance(blacklist, dict):
            raise TypeError("`blacklist` must be a dictionary.")
        for key, value in blacklist.items():
            if not isinstance(key, str):
                raise TypeError("Keys in `blacklist` must be strings.")
            if not isinstance(value, (list, tuple)):
                raise TypeError("Values in `blacklist` must be lists.")

    if xcats is None:
        if xcat_labels:
            raise ValueError("`xcat_labels` requires `xcats` to be defined.")
        xcats: List[str] = df["xcat"].unique().tolist()

    if cids is None:
        cids: List[str] = df["cid"].unique().tolist()

    df = reduce_df(
        df, xcats=xcats, cids=cids, start=start, end=end, blacklist=blacklist
    )

    if cumsum:
        df[val] = (
            df.sort_values(["cid", "xcat", "real_date"])[["cid", "xcat", val]]
            .groupby(["cid", "xcat"])
            .cumsum()
        )

    cross_mean_series: Optional[str] = f"mean_{xcats[0]}" if cs_mean else None
    if cs_mean:
        if len(xcats) > 1:
            raise ValueError("`cs_mean` cannot be True for multiple categories.")

        if len(cids) == 1:
            raise ValueError("`cs_mean` cannot be True for a single cross section.")

        df_mean: pd.DataFrame = (
            df.groupby(["real_date", "xcat"])[val].mean(numeric_only=True).reset_index()
        )

        df_mean["cid"] = "mean"
        df: pd.DataFrame = pd.concat([df, df_mean], axis=0)
        # Drop to save memory
        df_mean: pd.DataFrame = pd.DataFrame()

    if xcat_labels:
        # when `cs_mean` is True, `xcat_labels` may have one extra label
        if isinstance(xcat_labels, list):
            if len(xcat_labels) != len(xcats) and len(xcat_labels) != len(xcats) + int(
                cs_mean
            ):
                raise ValueError(
                    "`xcat_labels` must have same length as `xcats` "
                    "(or one extra label if `cs_mean` is True)."
                )
        elif isinstance(xcat_labels, dict):
            if not all([x in xcat_labels for x in xcats]):
                raise ValueError("Keys in `xcat_labels` must be a subset of `xcats`.")
            xcat_labels = [xcat_labels[x] for x in xcats if x in xcat_labels]
        

    if cid_labels:
        if isinstance(cid_labels, list):
            if len(cid_labels) != len(cids):
                raise ValueError("`cid_labels` must have same length as `cids`.")
        elif isinstance(cid_labels, dict):
            if not all([cid in cid_labels for cid in cids]):
                raise ValueError("Keys in `cid_labels` must be a subset of `cids`.")
        else:
            raise TypeError("`cid_labels` must be a list or a dictionary.")

    if cs_mean:
        if xcat_labels is None:
            xcat_labels = [xcats[0]]
            
        if len(xcat_labels) == 1:
            xcat_labels.append("Cross-Sectional Mean")

    facet_size: Optional[Tuple[float, float]] = (
        (aspect * height, height)
        if (aspect is not None and height is not None)
        else None
    )

    if xcat_grid and (len(xcats) == 1):
        xcat_grid: bool = False
        single_chart: bool = True

    if xcat_grid:
        if ncol > len(xcats):
            ncol: int = len(xcats)

        with FacetPlot(
            df=df,
            xcats=xcats,
            cids=cids,
            intersect=intersect,
            metrics=[val],
            tickers=[cross_mean_series] if cs_mean else None,
            start=start,
            end=end,
            blacklist=blacklist,
        ) as fp:
            fp.lineplot(
                share_y=same_y,
                share_x=not all_xticks,
                figsize=size,
                xcat_grid=True,  # Not to be confused with `xcat_grid` parameter
                # legend_labels=xcat_labels or None,
                facet_titles=xcat_labels or None,
                title=title,
                title_yadjust=title_adj,
                title_xadjust=title_xadj,
                compare_series=cross_mean_series if cs_mean else None,
                title_fontsize=title_fontsize,
                # title_fontsize=24,
                ncols=ncol,
                attempt_square=square_grid,
                facet_size=facet_size,
                legend_ncol=legend_ncol,
                legend_fontsize=legend_fontsize,
                interpolate=cumsum,
                ax_hline=ax_hline,
            )

    elif single_chart or (len(cids) == 1):
        with LinePlot(
            df=df,
            cids=cids,
            xcats=xcats,
            intersect=intersect,
            metrics=[val],
            tickers=[cross_mean_series] if cs_mean else None,
            start=start,
            end=end,
            blacklist=blacklist,
        ) as lp:
            lp.plot(
                metric=val,
                figsize=size,
                title=title,
                title_yadjust=title_adj,
                title_xadjust=title_xadj,
                compare_series=cross_mean_series if cs_mean else None,
                title_fontsize=title_fontsize,
                # title_fontsize=18,
                legend_ncol=legend_ncol,
                legend_fontsize=legend_fontsize,
                legend_labels=xcat_labels or None,
                ax_hline=ax_hline,
            )

    else:
        # Order cids by the values in cid_labels if provided, otherwise by the cids themselves
        if cid_labels and sort_cid_labels:
            if isinstance(cid_labels, list):
                cid_labels = {cid: label for cid, label in zip(cids, cid_labels)}
            cids = sorted(cids, key=lambda x: cid_labels.get(x, x))

        with FacetPlot(
            df=df,
            xcats=xcats,
            cids=cids,
            intersect=intersect,
            metrics=[val],
            tickers=[cross_mean_series] if cs_mean else None,
            start=start,
            end=end,
            blacklist=blacklist,
        ) as fp:
            show_legend: bool = True if cross_mean_series else False
            show_legend = show_legend or (len(xcats) > 1)
            if ncol > len(cids):
                ncol: int = len(cids)

            fp.lineplot(
                figsize=size,
                share_y=same_y,
                share_x=not all_xticks,
                title=title,
                # cid_xcat_grid=True,
                cid_grid=True,
                facet_titles=cid_labels or None,
                title_yadjust=title_adj,
                title_xadjust=title_xadj,
                compare_series=cross_mean_series if cs_mean else None,
                facet_size=facet_size,
                title_fontsize=title_fontsize,
                # title_fontsize=24,
                ncols=ncol,
                attempt_square=square_grid,
                legend=show_legend,
                legend_ncol=legend_ncol,
                legend_labels=xcat_labels,
                legend_fontsize=legend_fontsize,
                interpolate=cumsum,
                ax_hline=ax_hline,
            )


if __name__ == "__main__":
    from macrosynergy.visuals import FacetPlot
    from macrosynergy.management.simulate import make_test_df
    import numpy as np

    np.random.seed(42)

    cids: List[str] = [
        "USD",
        "EUR",
        "GBP",
        "AUD",
        "CAD",
        "JPY",
        "CHF",
        "NZD",
        "SEK",
        "NOK",
        "DKK",
        "INR",
    ]
    xcats: List[str] = [
        "FXXR",
        "EQXR",
        "RIR",
        "IR",
        "REER",
        "CPI",
        "PPI",
        "M2",
        "M1",
        "M0",
        "FXVOL",
        "FX",
    ]
    sel_cids: List[str] = ["GBP", "USD", "EUR"]
    sel_xcats: List[str] = ["FXXR", "EQXR", "RIR", "IR"]
    r_styles: List[str] = [
        "linear",
        "decreasing-linear",
        "sharp-hill",
        "sine",
        "four-bit-sine",
    ]
    df: pd.DataFrame = make_test_df(
        cids=list(set(cids) - set(sel_cids)),
        xcats=xcats,
        start="2000-01-01",
    )

    for rstyle, xcatx in zip(r_styles, sel_xcats):
        dfB: pd.DataFrame = make_test_df(
            cids=sel_cids,
            xcats=[xcatx],
            start="2000-01-01",
            style=rstyle,
        )
        df: pd.DataFrame = pd.concat([df, dfB], axis=0)

    for ix, cidx in enumerate(sel_cids):
        df.loc[df["cid"] == cidx, "value"] = (
            ((df[df["cid"] == cidx]["value"]) * (ix + 1)).reset_index(drop=True).copy()
        )

    for ix, xcatx in enumerate(sel_xcats):
        df.loc[df["xcat"] == xcatx, "value"] = (
            ((df[df["xcat"] == xcatx]["value"]) * (ix * 10 + 1))
            .reset_index(drop=True)
            .copy()
        )

    import time

    black = {
        "EUR": tuple(["2012-01-01", "2018-01-01"]),
        "GBP": ["2004-01-01", "2007-01-01"],
        "USD": ["2015-01-01", "2018-01-01"],
    }

    timer_start: float = time.time()
    timelines(
        df=df,
        xcats=sel_xcats,
        xcat_grid=False,
        square_grid=True,
        cids=sel_cids,
        blacklist=black,
        same_y=True,
        xcat_labels={"FXXR": "FX Returns", "EQXR": "Equity Returns", "RIR": "Real Interest Rate", "IR": "Interest Rate"},
    )

    # timelines(
    #     df=df,
    #     xcats=sel_xcats[0],
    #     cids=sel_cids,
    #     cs_mean=True,
    #     # xcat_grid=False,
    #     single_chart=True,
    #     blacklist=black,
    # )

    timelines(
        df=df,
        same_y=False,
        xcats=sel_xcats[0],
        cids=sel_cids,
        title=(
            "Plotting multiple cross sections for a single category \n with different "
            "y-axis!"
        ),
        # blacklist=black,
        cumsum=True,
        cid_labels={"USD": "Label 1", "EUR": "Label 2", "GBP": "Label 3"},
        xcat_labels={"FXXR": "Xcat Label"},
        cs_mean=True,
    )



from typing import List
import warnings

import pandas as pd
from datetime import datetime

from macrosynergy.management.types.qdf.classes import QuantamentalDataFrame
from macrosynergy.management.utils.df_utils import reduce_df


class BasePanelImputer:
    """
    Base class for imputing missing values in a panel DataFrame. Defines an overall
    structure for how the imputation should be performed, without the imputation method.
    Separate subclasses should be created for each imputation method, which will have a
    defined impute technique.

    Parameters
    ----------
    df : ~pandas.DataFrame
        DataFrame containing the panel data.
    xcats : List[str]
        List of extended categories.
    cids : List[str]
        List of cross sections.
    start : str
        Start date in ISO format.
    end : str
        End date in ISO format.
    min_cids : int
        Minimum number of cross sections required to perform imputation on a specific
        real date. Default is len(cids) // 2.
    postfix : str
        Postfix to add to the extended categories after imputation. Default is "F".
    """

    def __init__(
        self,
        df: pd.DataFrame,
        xcats: List[str],
        cids: List[str],
        start: str = None,
        end: str = None,
        min_cids: int = None,
        postfix: str = "F",
    ):
        if not isinstance(df, pd.DataFrame):
            raise TypeError("df must be a pandas DataFrame")
        else:
            df = QuantamentalDataFrame(df)
        if not isinstance(xcats, list):
            raise TypeError("xcats must be a list")
        if not isinstance(cids, list):
            raise TypeError("cids must be a list")
        if start is None:
            start = df["real_date"].min().strftime("%Y-%m-%d")
        if not isinstance(start, str):
            raise TypeError("start must be a string")
        else:
            try:
                start = datetime.strptime(start, "%Y-%m-%d")
            except ValueError:
                raise ValueError("start must be in the format 'YYYY-MM-DD'")
        if end is None:
            end = df["real_date"].max().strftime("%Y-%m-%d")
        if not isinstance(end, str):
            raise TypeError("end must be a string")
        else:
            try:
                end = datetime.strptime(end, "%Y-%m-%d")
            except ValueError:
                raise ValueError("end must be in the format 'YYYY-MM-DD'")
        if min_cids is None:
            min_cids = len(cids) // 2
        elif not isinstance(min_cids, int) or min_cids < 0:
            raise TypeError("min_cids must be a non-negative integer")
        if not isinstance(postfix, str):
            raise TypeError("postfix must be a string")

        if min_cids > df["cid"].nunique():
            raise ValueError(
                "min_cids must be less than or equal to the number of unique cids in df"
            )

        self.cids = cids
        self.xcats = xcats
        self.start = start
        self.end = end
        self.min_cids = min_cids
        self.postfix = postfix
        self.imputed = False
        self.blacklist = {xcat: {} for xcat in xcats}

        self.df = QuantamentalDataFrame(
            reduce_df(df, xcats=xcats, start=self.start, end=self.end).dropna()
        )
        self._as_categorical = df.InitializedAsCategorical

    def impute(self):
        """
        Returns the imputed DataFrame.
        """
        business_dates = (
            pd.date_range(start=self.start, end=self.end, freq="B")
            .strftime("%Y-%m-%d")
            .tolist()
        )

        full_idx = pd.MultiIndex.from_product(
            [business_dates, self.xcats, self.cids], names=["real_date", "xcat", "cid"]
        )

        imputed_df = pd.DataFrame(index=full_idx).reset_index()
        imputed_df["real_date"] = pd.to_datetime(imputed_df["real_date"])
        imputed_df = imputed_df.merge(
            self.df,
            how="outer",
            on=["real_date", "xcat", "cid"],
            suffixes=("", ""),
        )

        imputed_df["value"] = imputed_df.groupby(["real_date", "xcat"], observed=True)[
            "value"
        ].transform(self.get_impute_function)

        diff = self.df.merge(
            imputed_df, how="outer", on=["real_date", "xcat", "cid"], suffixes=("", "_")
        )
        diff["imputed"] = diff["value"].isnull() & ~diff["value_"].isnull()

        self.generate_blacklist(diff)

        if not self.imputed:
            warnings.warn(
                "No imputation was performed. Consider changing the impute_method or min_cids."
            )

        imputed_df = reduce_df(
            imputed_df, cids=self.cids, xcats=self.xcats, start=self.start, end=self.end
        )
        imputed_df["xcat"] = imputed_df["xcat"] + self.postfix
        imputed_df.dropna(inplace=True)
        return QuantamentalDataFrame(imputed_df, categorical=self._as_categorical)

    def get_impute_function(self, group):
        """
        Abstract method that should be implemented in a subclass. Defines the imputation
        technique to be used on a group of values.
        """
        raise NotImplementedError(
            "get_impute_function must be implemented in a subclass"
        )

    def generate_blacklist(self, diff: pd.DataFrame):
        """
        Generates a dictionary of cross sections and dates that have been imputed in the
        same format as a blacklist dictionary. For each cross section it stores a date
        range where imputation has been performed.

        Parameters
        ----------
        diff : ~pandas.DataFrame
            DataFrame containing the differences between the original and imputed
            DataFrames.
        """
        grouped = diff.groupby(["xcat", "cid"], observed=True)

        for (xcat, cid), group in grouped:
            group = group.sort_values(["real_date"]).reset_index(drop=True)
            imputed_group = group[group["imputed"]]
            if imputed_group.empty:
                continue

            consecutive_groups = (imputed_group.index.to_series().diff() != 1).cumsum()

            date_ranges = (
                imputed_group.groupby(consecutive_groups, observed=True)
                .agg(start=("real_date", "first"), end=("real_date", "last"))
                .reset_index(drop=True)
            )

            if len(date_ranges) == 1:
                self.blacklist[xcat][cid] = (
                    date_ranges.iloc[0]["start"],
                    date_ranges.iloc[0]["end"],
                )
            else:
                for i, row in date_ranges.iterrows():
                    self.blacklist[xcat][f"{cid}_{i+1}"] = (row["start"], row["end"])

    def return_blacklist(self, xcat: str = None):
        if not self.imputed:
            warnings.warn(
                "No imputation was performed. The blacklist is empty.", RuntimeWarning
            )
        return self.blacklist if xcat is None else self.blacklist[xcat]


class MeanPanelImputer(BasePanelImputer):
    """
    Imputer class that fills missing values with the global cross-sectional mean.
    If the group has less than min_cids non-missing values, the group is left as is.
    """

    def get_impute_function(self, group):
        if group.count() >= self.min_cids:
            self.imputed = True
            return group.fillna(group.mean())
        return group


class MedianPanelImputer(BasePanelImputer):
    """
    Imputer class that fills missing values with the global cross-sectional median.
    If the group has less than min_cids non-missing values, the group is left as is.
    """

    def get_impute_function(self, group):
        if group.count() >= self.min_cids:
            self.imputed = True
            return group.fillna(group.median())
        return group



from typing import Optional, List

from macrosynergy.management.types import QuantamentalDataFrame

from macrosynergy.panel.adjust_weights import adjust_weights


def linear_combination_adjustment(
    df: QuantamentalDataFrame,
    adj_zns_xcat: str,
    weights_xcat: str,
    cids: Optional[List[str]] = None,
    min_score: Optional[float] = None,
    coeff_new: Optional[float] = 0.5,
    normalize: bool = True,
    normalize_to_pct: bool = False,
    adj_name: str = "lincomb",
) -> QuantamentalDataFrame:
    """
    Adjust the weights of the zns scores based on the linear combination of the cross-sectional values.

    Parameters
    ----------
    df : QuantamentalDataFrame
        The input dataframe.

    zns_xcat : str
        The category of the zns scores to adjust. This category should be present in the `df`.

    min_score : float, optional
        The minimum score to consider. Default is None, where it is set to the minimum
        score discovered in the panel of `zns_xcat`.

    coeff_new : float, optional
        The coefficient to use for the new weights. Default is 0.5.

    normalize : bool, optional
        Whether to normalize the weights. Default is True.

    normalize_to_pct : bool, optional
        Whether to normalize the weights to percentages. Default is False.

    adj_name : str, optional
        The name of the new category. Default is "lincomb".

    Returns
    -------
    QuantamentalDataFrame
        The adjusted dataframe.
    """
    return adjust_weights(
        df=df,
        weights_xcat=weights_xcat,
        adj_zns_xcat=adj_zns_xcat,
        cids=cids,
        method="lincomb",
        params=dict(min_score=min_score, coeff_new=coeff_new),
        normalize=normalize,
        normalize_to_pct=normalize_to_pct,
        adj_name=adj_name,
    )


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_test_df

    df = make_test_df(
        xcats=["weights", "adj_zns", "other_xcat"],
        cids=["cid1", "cid2", "cid3", "cid4"],
    )

    df_res = linear_combination_adjustment(
        df,
        adj_zns_xcat="adj_zns",
        weights_xcat="weights",
        min_score=-3,
        coeff_new=0.5,
        adj_name="lincomb",
    )
    print(df_res)



"""
Function for calculating historic volatility of quantamental data.
"""

import numpy as np
import pandas as pd
from typing import List, Optional, Dict, Any
import warnings
from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import reduce_df, standardise_dataframe, get_eops
from macrosynergy.management.types import QuantamentalDataFrame


def historic_vol(
    df: pd.DataFrame,
    xcat: str = None,
    cids: List[str] = None,
    lback_periods: int = 21,
    lback_meth: str = "ma",
    half_life=11,
    start: str = None,
    end: str = None,
    est_freq: str = "D",
    blacklist: dict = None,
    remove_zeros: bool = True,
    postfix="ASD",
    nan_tolerance: float = 0.25,
):
    """
    Estimate historic annualized standard deviations of asset returns. The function can
    calculate the volatility using either a moving average or an exponential moving
    average method.

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized DataFrame with the following necessary columns: 'cid', 'xcat',
        'real_date' and 'value'. Will contain all of the data across all macroeconomic
        fields.
    xcat : str
        extended category denoting the return series for which volatility should be
        calculated. Note: in JPMaQS returns are represented in %, i.e. 5 means 5%.
    cids : List[str]
        cross sections for which volatility is calculated; default is all available for
        the category.
    lback_periods : int
        Number of lookback periods over which volatility is calculated. Default is 21.
    lback_meth : str
        Lookback method to calculate the volatility, Default is "ma". Alternative is
        "xma", Exponential Moving Average. Expects to receive either the aforementioned
        strings.
    half_life : int
        Refers to the half-time for "xma". Default is 11.
    start : str
        earliest date in ISO format. Default is None and earliest date in df is used.
    end : str
        latest date in ISO format. Default is None and latest date in df is used.
    est_freq : str
        Frequency of (re-)estimation of volatility. Options are 'D' for end of each day
        (default), 'W' for end of each work week, 'M' for end of each month, and 'Q' for end
        of each week.
    blacklist : dict
        cross sections with date ranges that should be excluded from the data frame. If
        one cross section has several blacklist periods append numbers to the cross section
        code.
    half_life : int
        Refers to the half-time for "xma" and full lookback period for "ma".
    remove_zeros : bool
        if True (default) any returns that are exact zeros will not be included in the
        lookback window and prior non-zero values are added to the window instead.
    postfix : str
        string appended to category name for output; default is "ASD".
    nan_tolerance : float
        maximum ratio of NaNs to non-NaNs in a lookback window, if exceeded the
        resulting volatility is set to NaN. Default is 0.25.

    Returns
    -------
    ~pandas.DataFrame
        standardized DataFrame with the estimated annualized standard deviations of the
        chosen category. If the input 'value' is in % (as is the standard in 
        JPMaQS) then the output will also be in %. 'cid', 'xcat', 'real_date' and 'value'.
    """

    df: QuantamentalDataFrame = QuantamentalDataFrame(df)
    est_freq = est_freq.lower()
    assert lback_meth in ["xma", "ma"], (
        "Lookback method must be either 'xma' "
        "(exponential moving average) or 'ma' (moving average)."
    )
    if lback_meth == "xma":
        assert (
            lback_periods > half_life
        ), "Half life must be shorter than lookback period."
        assert half_life > 0, "Half life must be greater than 0."
    assert est_freq in [
        "d",
        "w",
        "m",
        "q",
    ], "Estimation frequency must be one of 'D', 'W', 'M', or 'Q'."

    # assert nan tolerance is an int or float. must be >0. if >1 must be int
    assert isinstance(
        nan_tolerance, (int, float)
    ), "nan_tolerance must be an int or float."
    assert (
        0 <= nan_tolerance <= 1
    ), "nan_tolerance must be between 0.0 and 1.0 inclusive."

    df = reduce_df(
        df, xcats=[xcat], cids=cids, start=start, end=end, blacklist=blacklist
    )

    dfw = df.pivot(index="real_date", columns="cid", values="value")

    trigger_indices = get_eops(
        dates=pd.DataFrame(dfw.index),
        freq=est_freq,
    )

    def single_calc(
        row,
        dfw: pd.DataFrame,
        lback_periods: int,
        nan_tolerance: float,
        roll_func: callable,
        remove_zeros: bool,
        weights: Optional[np.ndarray] = None,
    ):
        """
        Helper function to calculate the historic volatility for a single row in the
        DataFrame.
        """
        target_dates = pd.bdate_range(end=row["real_date"], periods=lback_periods)
        target_df: pd.DataFrame = dfw.loc[dfw.index.isin(target_dates)]

        if weights is None:
            out = np.sqrt(252) * target_df.agg(roll_func, remove_zeros=remove_zeros)
        else:
            if len(weights) == len(target_df):
                out = np.sqrt(252) * target_df.agg(
                    roll_func, w=weights, remove_zeros=remove_zeros
                )
            else:
                return pd.Series(np.nan, index=target_df.columns)

        mask = (
            (
                target_df.isna().sum(axis=0)
                + (target_df == 0).sum(axis=0)
                + (lback_periods - len(target_df))
            )
            / lback_periods
        ) <= nan_tolerance
        # NOTE: dates with NaNs, dates with missing entries, and dates with 0s
        # are all treated as missing data and trigger a NaN in the output
        out[~mask] = np.nan

        return out

    expo_weights_arr: Optional[np.ndarray] = None
    if lback_meth == "xma":
        expo_weights_arr = expo_weights(lback_periods, half_life)

    if est_freq == "d":
        _args: Dict[str, Any] = dict(remove_zeros=remove_zeros)
        if lback_meth == "xma":
            _args["w"] = expo_weights_arr
            _args["func"] = expo_std
        else:
            _args["func"] = flat_std

        dfwa = np.sqrt(252) * dfw.rolling(window=lback_periods).agg(**_args)
    else:
        dfwa = pd.DataFrame(index=dfw.index, columns=dfw.columns)
        _args: Dict[str, Any] = dict(
            lback_periods=lback_periods,
            nan_tolerance=nan_tolerance,
            remove_zeros=remove_zeros,
        )
        if lback_meth == "xma":
            _args["weights"] = expo_weights_arr
            _args["roll_func"] = expo_std

        else:
            _args["roll_func"] = flat_std

        dfwa.loc[trigger_indices, :] = (
            dfwa.loc[trigger_indices, :]
            .reset_index(False)
            .apply(
                lambda row: single_calc(
                    row=row,
                    dfw=dfw,
                    **_args,
                ),
                axis=1,
            )
            .set_index(trigger_indices)
        )

        fills = {"d": 1, "w": 5, "m": 24, "q": 64}
        dfwa = dfwa.astype(float).reindex(dfw.index).ffill(limit=fills[est_freq])

    df_out = dfwa.unstack().reset_index().rename({0: "value"}, axis=1)

    # Create an initial mask for all rows to keep
    keep_mask = pd.Series(False, index=df_out.index)

    # Iterate over each cid and mark valid rows
    for cid in cids:
        # Get the date range for the current 'cid' in the original df
        loc_bools = df["cid"] == cid
        if df[loc_bools].empty:
            warnings.warn(f"No data for {cid}_{xcat}. Skipping.")
            continue
        min_date = df.loc[loc_bools, "real_date"].min()
        max_date = df.loc[loc_bools, "real_date"].max()

        # Generate valid date range for the current 'cid'
        valid_dates = pd.bdate_range(start=min_date, end=max_date)

        # Update the keep_mask for rows corresponding to current 'cid' with valid dates
        sel_bools = df_out["cid"] == cid
        sel_dts = df_out["real_date"].isin(valid_dates)

        keep_mask |= sel_bools & sel_dts

    # Apply the mask to df_out
    df_out = df_out[keep_mask].reset_index(drop=True)

    df_out = QuantamentalDataFrame.from_long_df(
        df=df_out,
        xcat=xcat + postfix,
        categorical=df.InitializedAsCategorical,
    )
    return standardise_dataframe(df_out)


def expo_weights(lback_periods: int = 21, half_life: int = 11):
    """
    Calculates exponential series weights for finite horizon, normalized to 1.

    Parameters
    ----------
    lback_periods : int
        Number of lookback periods over which volatility is calculated. Default is 21.
    half_life : int
        Refers to the half-time for "xma" and full lookback period for "ma". Default is
        11.

    Returns
    -------
    ~numpy.ndarray
        An Array of weights determined by the length of the lookback period.

    Notes
    -----
    50% of the weight allocation will be applied to the number of days delimited by the
    half_life.
    """

    decf = 2 ** (-1 / half_life)
    weights = (1 - decf) * np.array(
        [decf ** (lback_periods - ii - 1) for ii in range(lback_periods)]
    )
    weights = weights / sum(weights)

    return weights


def expo_std(x: np.ndarray, w: np.ndarray, remove_zeros: bool = True):
    """
    Estimate standard deviation of returns based on exponentially weighted absolute
    values.

    Parameters
    ----------
    x : ~numpy.ndarray
        array of returns
    w : ~numpy.ndarray
        array of exponential weights (same length as x); will be normalized to 1.
    remove_zeros : bool
        removes zeroes as invalid entries and shortens the effective window.

    Returns
    -------
    float
        exponentially weighted mean absolute value (as proxy of return standard
        deviation).
    """

    assert len(x) == len(w), "weights and window must have same length"
    if remove_zeros:
        x = x[x != 0]
        w = w[0 : len(x)] / sum(w[0 : len(x)])
    w = w / sum(w)  # weights are normalized
    mabs = np.sum(np.multiply(w, np.abs(x)))
    return mabs


def flat_std(x: np.ndarray, remove_zeros: bool = True):
    """
    Estimate standard deviation of returns based on exponentially weighted absolute
    values.

    Parameters
    ----------
    x : ~numpy.ndarray
        array of returns
    remove_zeros : bool
        removes zeroes as invalid entries and shortens the effective window.

    Returns
    -------
    float
        flat weighted mean absolute value (as proxy of return standard deviation).
    """

    if remove_zeros:
        x = x[x != 0]
    mabs = np.mean(np.abs(x))
    return mabs


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )

    df_cids.loc["AUD"] = ["2010-01-01", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD"] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-10-30", -0.2, 0.5]
    df_cids.loc["USD"] = ["2013-01-01", "2020-09-30", -0.2, 0.5]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR"] = ["2010-01-01", "2020-12-31", 0, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2011-01-01", "2020-10-30", 1, 2, 0.9, 0.5]
    df_xcats.loc["GROWTH"] = ["2012-01-01", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2013-01-01", "2020-10-30", 1, 2, 0.8, 0.5]
    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])

    print("Calculating historic volatility with the moving average method")
    df = historic_vol(
        dfd,
        cids=cids,
        xcat="XR",
        lback_periods=7,
        lback_meth="ma",
        est_freq="w",
        half_life=3,
        remove_zeros=True,
    )

    print(df.head(10))

    print("Calculating historic volatility with the exponential moving average method")
    df = historic_vol(
        dfd,
        cids=cids,
        xcat="XR",
        lback_periods=7,
        lback_meth="xma",
        est_freq="w",
        half_life=3,
        remove_zeros=True,
    )

    print(df.head(10))



"""
A set of tools for cross-validation of panel data.
"""

import numpy as np
import pandas as pd
import datetime

from macrosynergy.learning import ExpandingKFoldPanelSplit
from macrosynergy.learning.splitters.base_splitters import BasePanelSplit

from sklearn.model_selection import cross_validate
from sklearn.metrics import make_scorer


def panel_cv_scores(
    X,
    y,
    splitter,
    estimators,
    scoring,
    show_longbias = True,
    show_std = False,
    verbose = 1,
    n_jobs = -1,
):
    """
    Returns a dataframe of cross-validation scores for a collection of models, with
    respect to a cross-validation splitter and a set of scorers.

    Parameters
    ----------
    X : pd.DataFrame
        Input feature matrix.
    y : pd.DataFrame or pd.Series
        Target variable.
    splitter : BasePanelSplit
        Panel cross-validation splitter.
    estimators : dict
        Dictionary of models.
    scoring : dict
        Dictionary of scorers.
    show_longbias : bool, optional, default=True
        Whether to show the proportion of times a model predicts a positive return.
    show_std : bool, optional, default=False
        Whether to show the standard deviation of the cross-validation scores over folds.
    verbose : int, optional, default=1
        Verbosity level.
    n_jobs : int, optional, default=-1
        Number of jobs to run in parallel.

    Returns
    -------
    pd.DataFrame
        Dataframe of cross-validation scores.

    Notes
    -----
    This function returns a dataframe that is multi-indexed with the outer index
    representing a metric and the inner index representing the mean & (optionally) a
    standard deviation over validation splits. The columns are the estimators.
    """

    # check input types
    if not isinstance(X, pd.DataFrame):
        raise TypeError("X must be a pandas dataframe.")
    if not isinstance(y, (pd.DataFrame, pd.Series)):
        raise TypeError("y must be a pandas dataframe or series.")
    if not isinstance(X.index, pd.MultiIndex):
        raise TypeError("X must be multi-indexed.")
    if not isinstance(y.index, pd.MultiIndex):
        raise TypeError("y must be multi-indexed.")
    if not isinstance(splitter, BasePanelSplit):
        raise TypeError("splitter must be an inherit from BasePanelSplit.")
    if not isinstance(estimators, dict):
        raise TypeError("estimators must be a dictionary.")
    if estimators == {}:
        raise ValueError("estimators must not be an empty dictionary.")
    if np.any([not isinstance(est_name, str) for est_name in estimators.keys()]):
        raise TypeError("estimator names must all be strings.")
    if np.any([not isinstance(est, object) for est in estimators.values()]):
        raise TypeError("estimators must all be objects.")
    if not isinstance(scoring, dict):
        raise TypeError("scoring must be a dictionary.")
    if scoring == {}:
        raise ValueError("scoring must not be an empty dictionary.")
    if np.any([not isinstance(metric_name, str) for metric_name in scoring.keys()]):
        raise TypeError("scorer names must all be strings.")

    if not isinstance(show_longbias, bool):
        raise TypeError("show_longbias must be a boolean.")
    if not isinstance(show_std, bool):
        raise TypeError("show_std must be a boolean.")
    if not isinstance(verbose, int):
        raise TypeError("verbose must be an integer.")
    if not isinstance(n_jobs, int):
        raise TypeError("n_jobs must be an integer.")

    # check the dataframes are in the right format
    if not isinstance(X.index.get_level_values(1)[0], datetime.date):
        raise TypeError("The inner index of X must be datetime.date.")
    if not isinstance(y.index.get_level_values(1)[0], datetime.date):
        raise TypeError("The inner index of y must be datetime.date.")
    if not X.index.equals(y.index):
        raise ValueError(
            "The indices of the input dataframe X and the output dataframe y don't match."
        )

    # check that there is at least one estimator and at least one scoring metric
    if len(estimators) <= 0:
        raise ValueError("There must be at least one estimator provided.")
    if len(scoring) <= 0:
        raise ValueError("There must be at least one scoring metric provided.")
    if verbose < 0:
        raise ValueError("verbose must be a non-negative integer.")
    if (n_jobs < 1) & (n_jobs != -1):
        raise ValueError("n_jobs must either be a positive integer or equal to -1.")

    # construct the dataframe to return
    if show_longbias:
        scoring = scoring.copy()
        scoring["Positive prediction ratio"] = make_scorer(
            lambda y_true, y_pred: np.sum(y_pred > 0) / len(y_pred)
        )
        scoring["Positive test-target ratio"] = make_scorer(
            lambda y_true, y_pred: np.sum(y_true > 0) / len(y_true)
        )

    results: dict = {
        estimator_name: {} for estimator_name in estimators
    }

    for estimator_name, estimator in estimators.items():
        if verbose:
            print(f"Calculating walk-forward validation metrics for {estimator_name}.")
        cv_results: dict = cross_validate(
            estimator,
            X,
            y,
            cv=splitter,
            scoring=scoring,
            verbose=verbose,
            n_jobs=n_jobs,
        )

        for metric_name in scoring:
            score: np.ndarray = cv_results[f"test_{metric_name}"]
            results[estimator_name][metric_name] = np.mean(score)
            if show_std:
                results[estimator_name][f"{metric_name}_std"] = np.std(score)

    metrics_df = pd.DataFrame(results)
    if show_std:
        multi_index: pd.MultiIndex = pd.MultiIndex.from_tuples(
            [
                (key.split("_")[0], "std" if "std" in key else "mean")
                for key in metrics_df.index
            ]
        )
        metrics_df.index = multi_index

    return metrics_df


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_qdf
    import macrosynergy.management as msm
    import macrosynergy.learning as msl

    from sklearn.linear_model import LinearRegression, Lasso
    from sklearn.metrics import (
        root_mean_squared_error,
        mean_absolute_error,
        mean_absolute_percentage_error,
    )

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example 1: Unbalanced panel """

    df_cids2 = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids2.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids2.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids2.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids2.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats2 = pd.DataFrame(index=xcats, columns=cols)
    df_xcats2.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats2.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats2.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats2.loc["INFL"] = ["2000-01-01", "2020-12-31", 1, 2, 0.8, 0.5]

    dfd2 = make_qdf(df_cids2, df_xcats2, back_ar=0.75)
    dfd2["grading"] = np.ones(dfd2.shape[0])
    black = {"GBP": ["2009-01-01", "2012-06-30"], "CAD": ["2018-01-01", "2100-01-01"]}
    dfd2 = msm.reduce_df(df=dfd2, cids=cids, xcats=xcats, blacklist=black)

    dfd2 = dfd2.pivot(index=["cid", "real_date"], columns="xcat", values="value")
    X2 = dfd2.drop(columns=["XR"])
    y2 = dfd2["XR"]

    # 1) Demonstration of panel_cv_scores
    splitex = ExpandingKFoldPanelSplit(n_splits=100)
    models = {"OLS": LinearRegression(), "Lasso": Lasso(alpha=0.05)}
    metrics = {
        "neg_rmse": make_scorer(root_mean_squared_error, greater_is_better=False),
        "neg_mae": make_scorer(mean_absolute_error, greater_is_better=False),
        "neg_mape": make_scorer(mean_absolute_percentage_error, greater_is_better=False),
        "acc": make_scorer(msl.regression_accuracy),
        "bac": make_scorer(msl.regression_balanced_accuracy),
        "map": make_scorer(msl.panel_significance_probability),
        "sharpe": make_scorer(msl.sharpe_ratio),
        "sortino": make_scorer(msl.sortino_ratio),
    }
    df_ev = panel_cv_scores(
        X2,
        y2,
        splitter=splitex,
        estimators=models,
        scoring=metrics,
        show_longbias=True,
        show_std=False,
        n_jobs=-1,
        verbose=1,
    )
    print(df_ev)



"""
Implementation of the ProxyPnL class.
"""

import pandas as pd
from numbers import Number
from typing import List, Union, Tuple, Optional

from macrosynergy.management.utils import (
    reduce_df,
    is_valid_iso_date,
)
from macrosynergy.management.types import QuantamentalDataFrame, NoneType
import macrosynergy.visuals as msv
from macrosynergy.pnl import (
    notional_positions,
    contract_signals,
    proxy_pnl_calc,
    TransactionCosts,
)


class ProxyPnL(object):
    """
    The purpose of this class is to facilitate PnL estimation under the consideration of
    AUM, volatility targeting or leverage, and transaction costs. The class is designed
    to be used in a step-by-step manner, where the user first contracts signals, then
    calculates notional positions, and finally calculates the proxy PnL.

    The steps for generating the PnL are as follows:
    - Contract signals: Contract signals for the given contracts and contract types.
    - Notional positions: Calculate notional (dollar) positions for the given contract
        signals.
    - Proxy PnL calculation: Calculate the proxy PnL and transaction costs for the given
        notional positions.

    Parameters
    ----------
    df : QuantamentalDataFrame
        DataFrame containing the data to be used in the PnL estimation. Initially, this
        DataFrame should contain the data used to contract signals (i.e. raw signals).
    transaction_costs_object : TransactionCosts
        Object containing the transaction costs data.
    start : str, optional
        Start date for the PnL estimation. If not provided, the minimum date in the
        DataFrame is used.
    end : str, optional
        End date for the PnL estimation. If not provided, the maximum date in the
        DataFrame is used.
    blacklist : dict, optional
        The blacklist dictionary to be applied to the input data.
    rstring : str, optional
        A string used to specify the returns to be used in the PnL estimation.
    portfolio_name : str, optional
        The name given to the (current) portfolio. In the return outputs, the portfolio
        name is used to identify and aggregate the PnL and transaction costs.
    sname : str, optional
        The name given to the strategy,
    pname : str, optional
        The name given to the positions.

    """

    def __init__(
        self,
        df: QuantamentalDataFrame,
        transaction_costs_object: Optional[TransactionCosts],
        start: Optional[str] = None,
        end: Optional[str] = None,
        blacklist: Optional[dict] = None,
        rstring: str = "XR",
        portfolio_name: str = "GLB",
        sname: str = "STRAT",
        pname: str = "POS",
    ):
        self.sname = sname
        self.portfolio_name = portfolio_name
        self.pname = pname
        self.blacklist = blacklist
        self.cs_df = None
        self.npos_df = None
        self.df = reduce_df(df=QuantamentalDataFrame(df), blacklist=blacklist)
        self.start = start or df["real_date"].min().strftime("%Y-%m-%d")
        self.end = end or df["real_date"].max().strftime("%Y-%m-%d")
        self.rstring = rstring
        self.transaction_costs_object: Optional[TransactionCosts] = None
        if not all(map(is_valid_iso_date, [self.start, self.end])):
            raise ValueError(f"Invalid date format: {self.start}, {self.end}")

        if transaction_costs_object is None:
            pass  # allowed for no-transaction-costs case
        elif isinstance(transaction_costs_object, TransactionCosts):
            transaction_costs_object.check_init()
            self.transaction_costs_object: TransactionCosts = transaction_costs_object
        else:
            raise ValueError(
                "Invalid type for `transaction_costs_object`. Expected `TransactionCosts` object."
            )

        assert hasattr(
            self, "transaction_costs_object"
        ), "Failed to initialize `self.transaction_costs_object`"

    def contract_signals(
        self,
        sig: str,
        cids: List[str],
        ctypes: List[str],
        cscales: Optional[List[Union[Number, str]]] = None,
        csigns: Optional[List[int]] = None,
        hbasket: Optional[List[str]] = None,
        hscales: Optional[List[Union[Number, str]]] = None,
        hratios: Optional[str] = None,
        blacklist: Optional[dict] = None,
        *args,
        **kwargs,
    ) -> QuantamentalDataFrame:
        """
        Contract signals for the given contracts and contract types.
        The method uses the same dataframe as the one used to initialize the class.
        The function stores the contract signals DataFrame as an attribute of the class
        (`self.cs_df`), and also returns the same DataFrame for convenience.

        See :func:`macrosynergy.pnl.contract_signals` for more information on the other
        parameters.

        Returns
        -------
        QuantamentalDataFrame
        """
        self.fids = [f"{cid}_{ctype}" for cid in cids for ctype in ctypes]
        cs_df: QuantamentalDataFrame = contract_signals(
            df=self.df,
            sig=sig,
            cids=cids,
            ctypes=ctypes,
            cscales=cscales,
            csigns=csigns,
            hbasket=hbasket,
            hscales=hscales,
            hratios=hratios,
            start=self.start,
            end=self.end,
            blacklist=blacklist or self.blacklist,
            sname=self.sname,
            *args,
            **kwargs,
        )
        self.cs_df: QuantamentalDataFrame = cs_df
        return cs_df

    def notional_positions(
        self,
        df: QuantamentalDataFrame = None,
        sname: str = None,
        fids: List[str] = None,
        aum: Number = 100,
        dollar_per_signal: Number = 1.0,
        slip: int = 1,
        leverage: Optional[Number] = None,
        vol_target: Optional[Number] = None,
        nan_tolerance: float = 0.25,
        remove_zeros: bool = True,
        rebal_freq: str = "m",
        lback_meth: str = "ma",
        est_freqs: Union[str, List[str]] = ["D", "W", "M"],
        est_weights: Union[Number, List[Number]] = [1, 1, 1],
        lback_periods: Union[int, List[int]] = [-1, -1, -1],
        half_life: Union[int, List[int]] = [11, 5, 6],
        rstring: str = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        blacklist: Optional[dict] = None,
        pname: str = "POS",
    ) -> Union[
        QuantamentalDataFrame,
        Tuple[QuantamentalDataFrame, QuantamentalDataFrame],
        Tuple[QuantamentalDataFrame, pd.DataFrame],
        Tuple[QuantamentalDataFrame, QuantamentalDataFrame, pd.DataFrame],
    ]:
        """
        Calculate notional positions for the given contract signals.
        The method uses the contract signals calculated in the previous step. The user
        may additionally provide more data that may be used as a new dataframe.

        The method stores the notional positions DataFrame, the portfolio volatility
        DataFrame, and the variance-covariance matrix DataFrame as attributes of the
        class (`self.npos_df`, `self.pvol_df`, and `self.vcv_df`, respectively). It also
        returns the notional positions DataFrame for convenience.

        See :func:`macrosynergy.pnl.notional_positions` for more information on the other
        parameters.

        Returns
        -------
        QuantamentalDataFrame
            The notional positions DataFrame
        """

        fids = fids or self.fids
        if df is None:
            if hasattr(self, "cs_df") and self.cs_df is not None:
                df = self.cs_df
            else:
                raise ValueError(
                    "Either pass a DataFrame with contract signals "
                    "or run `ProxyPnL.contract_signals` first."
                )
        sname = sname or self.sname
        start = start or self.start
        end = end or self.end
        blacklist = blacklist or self.blacklist
        rstring = rstring or self.rstring

        outs: Union[
            Tuple[QuantamentalDataFrame, QuantamentalDataFrame, pd.DataFrame],
            QuantamentalDataFrame,
        ] = notional_positions(
            df=pd.concat((self.df, df), axis=0),
            sname=sname,
            fids=fids,
            aum=aum,
            dollar_per_signal=dollar_per_signal,
            slip=slip,
            leverage=leverage,
            vol_target=vol_target,
            nan_tolerance=nan_tolerance,
            remove_zeros=remove_zeros,
            rebal_freq=rebal_freq,
            lback_meth=lback_meth,
            est_freqs=est_freqs,
            est_weights=est_weights,
            lback_periods=lback_periods,
            half_life=half_life,
            rstring=rstring,
            start=start,
            end=end,
            blacklist=blacklist,
            pname=pname,
            return_pvol=True,
            return_vcv=True,
        )
        if isinstance(outs, QuantamentalDataFrame):
            assert isinstance(outs, QuantamentalDataFrame)
            outs = (outs, None, None)  # to avoid multiple flow control
        assert len(outs) == 3
        assert isinstance(outs[0], QuantamentalDataFrame)
        assert isinstance(outs[1], (QuantamentalDataFrame, NoneType))
        assert isinstance(outs[2], (pd.DataFrame, NoneType))

        self.npos_df: QuantamentalDataFrame = outs[0]
        self.pvol_df: QuantamentalDataFrame = outs[1]
        self.vcv_df: QuantamentalDataFrame = outs[2]
        outs = None
        return self.npos_df

    def proxy_pnl_calc(
        self,
        spos: str = None,
        portfolio_name: str = None,
        df: QuantamentalDataFrame = None,
        roll_freqs: Optional[dict] = None,
        rstring: str = None,
        pnl_name: str = "PNL",
        tc_name: str = "TCOST",
    ) -> Union[QuantamentalDataFrame, Tuple[QuantamentalDataFrame, ...]]:
        """
        Calculate the proxy PnL and transaction costs for the given notional positions.
        The method uses the notional positions calculated in the previous step. The user
        may additionally provide more data that may be used as a new dataframe.

        The method stores the proxy PnL DataFrame, the transaction costs DataFrame, and
        the proxy PnL excluding costs DataFrame as attributes of the class (`self.proxy_pnl`,
        `self.txn_costs_df`, and `self.pnl_excl_costs`, respectively). It also returns the
        proxy PnL DataFrame for convenience.

        See :func:`macrosynergy.pnl.proxy_pnl_calc` for more information on the other
        parameters.

        Returns
        -------
        QuantamentalDataFrame
            The proxy PnL DataFrame.
        """
        if df is None:
            if hasattr(self, "npos_df") and self.npos_df is not None:
                df = self.npos_df
            else:
                raise ValueError(
                    "Either pass a DataFrame with notional positions "
                    "or run `ProxyPnL.notional_positions` (and `contract_signals`) first."
                )
        spos: str = spos or (self.sname + "_" + self.pname)
        portfolio_name: str = portfolio_name or self.portfolio_name
        rstring: str = rstring or self.rstring

        outs: Tuple[QuantamentalDataFrame, ...] = proxy_pnl_calc(
            df=pd.concat((self.df, df), axis=0),
            transaction_costs_object=self.transaction_costs_object,
            spos=spos,
            rstring=rstring,
            portfolio_name=portfolio_name,
            roll_freqs=roll_freqs,
            start=self.start,
            end=self.end,
            blacklist=self.blacklist,
            pnl_name=pnl_name,
            tc_name=tc_name,
            return_pnl_excl_costs=True,
            return_costs=True,
        )
        assert len(outs) == 3
        assert all(map(lambda x: isinstance(x, QuantamentalDataFrame), outs))
        self.proxy_pnl: QuantamentalDataFrame = outs[0]
        self.txn_costs_df: QuantamentalDataFrame = outs[1]
        self.pnl_excl_costs: QuantamentalDataFrame = outs[2]
        outs = None

        return self.proxy_pnl

    def plot_pnl(self, title: str = "Proxy PnL", cumsum: bool = True, **kwargs):
        """
        Plot the proxy PnL DataFrame. The method uses the proxy PnL calculated in the
        previous step.

        Parameters
        ----------
        title : str, optional
            Title of the plot.
        cumsum : bool, optional
            Whether to plot the cumulative sum of the proxy PnL.
        kwargs
            Additional keyword arguments to be passed to the `timelines` function.
            See :func:`macrosynergy.visuals.timelines` for more information.
        """
        cdf = pd.concat((self.proxy_pnl, self.pnl_excl_costs), axis=0)
        rdf = reduce_df(cdf, cids=["GLB"])
        msv.timelines(rdf, title=title, cumsum=cumsum)


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_test_df

    cids_dmfx = ["CHF", "SEK", "NOK", "CAD", "GBP", "NZD", "JPY", "AUD"]
    fxblack = {"CHF": ("2011-10-03 00:00:00", "2015-01-30 00:00:00")}

    xcats = ["FX", "IRS", "CDS"]
    dfx = make_test_df(cids=cids_dmfx, xcats=xcats)
    txn_obj = TransactionCosts.download(verbose=True)

    p = ProxyPnL(
        df=dfx,
        transaction_costs_object=txn_obj,
        blacklist=fxblack,
        start="2001-01-01",
        end="2020-01-01",
        rstring="XR_NSA",
    )
    p.contract_signals(
        sig="CPIXFE_SJA_P6M6ML6ARvIETvBMZN",
        cids=cids_dmfx,
        ctypes=["FX"],
        cscales=["FXXRxLEV10_NSA"],
        relative_value=False,
        hbasket=["EUR_FX"],  # TODO invert asset class or returns?
        hscales=["FXXRxLEV10_NSA"],
        hratios="FXEURBETA",
    )
    p.notional_positions(
        aum=100,
        vol_target=10,
        rebal_freq="m",
        slip=1,
        est_freqs=["D", "W", "M"],
        est_weights=[1, 1, 1],
        lback_periods=[-1, -1, -1],
        lback_meth="xma",
        half_life=[11, 5, 6],
    )
    p.proxy_pnl_calc()



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin, MetaEstimatorMixin

class DataFrameTransformer(BaseEstimator, TransformerMixin, MetaEstimatorMixin):
    """
    Meta estimator to reconvert a transformed numpy array back to a multiindexed 
    pandas DataFrame. This maintains the multi-indexed panel structure.

    Parameters
    ----------
    transformer : TransformerMixin
        A scikit-learn transformer with a fit and transform method.
        
    Notes
    -----
    Many scikit-learn compatible transformers convert pandas DataFrames to numpy arrays.
    This can be problematic when working with panel models that require knowledge of the
    panel structure. This class wraps around such transformers to ensure that the output
    is a pandas DataFrame, preserving the original index.

    When no column names are provided, default names of the form "Factor_0", "Factor_1", etc.
    are used for the transformed DataFrame. If column names are provided, they will be used
    instead.
    """
    def __init__(self, transformer, column_names=None):
        # Checks
        if not isinstance(transformer, TransformerMixin):
            raise TypeError("transformer must be a scikit-learn transformer.")
        if column_names is not None and not isinstance(column_names, list):
            raise TypeError("column_names must be a list of strings or None.")
        if column_names is not None and len(column_names) == 0:
            raise ValueError("column_names cannot be an empty list.")
        if column_names is not None and not all(isinstance(name, str) for name in column_names):
            raise ValueError("All column names must be strings.")
        if column_names is not None and len(set(column_names)) != len(column_names):
            raise ValueError("All column names must be unique.")
        
        # Attributes
        self.transformer = transformer
        self.column_names = column_names

    def fit(self, X, y=None):
        """
        Fit the underlying transformer.

        Parameters
        ----------
        X : pd.DataFrame
            Pandas dataframe of input features.
        y : pd.Series or pd.DataFrame or np.ndarray
            Pandas series, dataframe or numpy array of targets associated with each sample
            in X.
        """
        # Checks
        self._check_fit_params(X, y)

        # Fit estimator
        self.transformer.fit(X, y)

        return self

    def transform(self, X):
        """
        Transform the input data based on the underlying transformer, but return a 
        pandas DataFrame instead of a numpy array.

        Parameters
        ----------
        X : pd.DataFrame or numpy array
            Input feature matrix.

        Returns
        -------
        pd.DataFrame
            Transformed data as a pandas DataFrame, preserving the original index and
            using either provided column names or default names.
        """
        # Checks
        self._check_predict_params(X)

        # Transform the data
        transformation = self.transformer.transform(X)

        # Check the number of column names provided is more than or equal to the number of features
        if self.column_names is not None and len(self.column_names) < transformation.shape[1]:
            raise ValueError(
                "The number of column names provided must be greater than or equal to the "
                "number of features in the transformed data."
            )

        # Return the data in the correct format
        if isinstance(transformation, pd.DataFrame):
            return transformation
        else:
            # scikit-learn returns a numpy array, convert it back to DataFrame
            if self.column_names is None:
                columns = [f"Factor_{i}" for i in range(transformation.shape[1])]
            else:
                columns = self.column_names[:transformation.shape[1]]

            return pd.DataFrame(
                data = transformation,
                columns = columns,
                index = X.index,
            )
        
    def __getattr__(self, name):
        """
        Delegate attribute access to the underlying transformer.

        Parameters
        ----------
        name : str
            The name of the attribute to access.
        """
        # Precent infinite recursion
        if name == "transformer":
            raise AttributeError()
        return getattr(self.transformer, name)
    
    def _check_fit_params(self, X, y):
        """
        Checks for fit method parameters.
        """
        # X
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the dataframe transformer must be a pandas "
                "dataframe."
            )
        # y
        if not isinstance(y, (pd.Series, pd.DataFrame, np.ndarray)):
            raise TypeError(
                "Target vector for the dataframe transformer must be either a pandas series, "
                "dataframe or numpy array."
            )
        if isinstance(y, pd.DataFrame):
            if y.shape[1] != 1:
                raise ValueError(
                    "The dependent variable dataframe must have only one column. If used "
                    "as part of an sklearn pipeline, ensure that previous steps return "
                    "a pandas series or dataframe."
                )
        if isinstance(y, np.ndarray):
            if y.ndim != 1:
                raise ValueError(
                    "When the target vector for the probability estimator is a numpy "
                    "array, it must have one dimension."
                )
        if X.shape[0] != y.shape[0]:
            raise ValueError(
                "The number of samples in the input feature matrix must match the number "
                "of samples in the target vector."
            )
        # Check NaN values
        if X.isnull().values.any():
            raise ValueError(
                "The input feature matrix contains NaN values. Please handle missing "
                "values before fitting the transformer."
            )
        if isinstance(y, (pd.DataFrame, pd.Series)) and y.isnull().values.any():
            raise ValueError(
                "The target vector contains NaN values. Please handle missing values before "
                "fitting the transformer."
            )
        elif isinstance(y, np.ndarray) and np.isnan(y).any():
            raise ValueError(
                "The target vector contains NaN values. Please handle missing values before "
                "fitting the transformer."
            )
        
        # Check that the number of column names provided is equal to the number of features
        if self.column_names is not None and len(self.column_names) != X.shape[1]:
            raise ValueError(
                "The number of column names provided must match the number of features."
            )

    def _check_predict_params(self, X):
        """
        Checks for predict method parameters.
        """
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the probability estimator must be a pandas "
                "dataframe."
            )

        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                "The number of features in the input feature matrix must match the number "
                "seen in training."
            )
        
        # Check that the number of column names provided equal to the number of features
        if self.column_names is not None and len(self.column_names) != X.shape[1]:
            raise ValueError(
                "The number of column names provided must be equal to the number of features."
            )
        
if __name__ == "__main__":
    import macrosynergy.management as msm
    from macrosynergy.management.simulate import make_qdf
    import pandas as pd
    import numpy as np

    from sklearn.decomposition import PCA
    from sklearn.pipeline import Pipeline


    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {
        "GBP": (
            pd.Timestamp(year=2009, month=1, day=1),
            pd.Timestamp(year=2012, month=6, day=30),
        ),
        "CAD": (
            pd.Timestamp(year=2015, month=1, day=1),
            pd.Timestamp(year=2100, month=1, day=1),
        ),
    }

    train = msm.categories_df(
        df=dfd, xcats=xcats, cids=cids, val="value", blacklist=black, freq="M", lag=1
    ).dropna()

    # Training set
    X_train = train.drop(columns=["XR"])
    y_train = np.sign(train["XR"])

    # Model fit
    print(DataFrameTransformer(PCA(n_components = 2)).fit_transform(X_train, y_train))
    print(DataFrameTransformer(PCA(n_components = 2), column_names = [f"PCA_{i}" for i in range(1, 4)]).fit_transform(X_train, y_train))



from .splitters import (
    ExpandingKFoldPanelSplit,
    RollingKFoldPanelSplit,
    RecencyKFoldPanelSplit,
    ExpandingIncrementPanelSplit,
    ExpandingFrequencyPanelSplit,
)
from .cv_tools import panel_cv_scores

from .preprocessing import (
    LassoSelector,
    LarsSelector,
    MapSelector,
    ZnScoreAverager,
    PanelMinMaxScaler,
    PanelStandardScaler,
    BasePanelScaler,
    BasePanelSelector,
    PanelPCA,
)
from .model_evaluation import (
    neg_mean_abs_corr,
    panel_significance_probability,
    sharpe_ratio,
    sortino_ratio,
    regression_accuracy,
    regression_balanced_accuracy,
    correlation_coefficient,
    create_panel_metric,
    regression_mcc,
)
from .sequential import SignalOptimizer, BetaEstimator, ReturnForecaster

from .forecasting import (
    LADRegressor,
    NaiveRegressor,
    SignWeightedLinearRegression,
    TimeWeightedLinearRegression,
    SignWeightedLADRegressor,
    TimeWeightedLADRegressor,
    ModifiedLinearRegression,
    ModifiedSignWeightedLinearRegression,
    ModifiedTimeWeightedLinearRegression,
    LinearRegressionSystem,
    LADRegressionSystem,
    RidgeRegressionSystem,
    CorrelationVolatilitySystem,
    KNNClassifier,
    ProbabilityEstimator,
    VotingRegressor,
    VotingClassifier,
    FIExtractor,
    DataFrameTransformer,
    GlobalLocalRegression,
    CountryByCountryRegression,
)

from .random_effects import RandomEffects

__all__ = [
    # splitters
    "ExpandingKFoldPanelSplit",
    "RollingKFoldPanelSplit",
    "RecencyKFoldPanelSplit",
    "ExpandingIncrementPanelSplit",
    "ExpandingFrequencyPanelSplit",
    "BasePanelSplit",
    # cv_tools
    "panel_cv_scores",
    # preprocessing
    "BasePanelSelector",
    "LassoSelector",
    "LarsSelector",
    "MapSelector",
    # transformers
    "BasePanelScaler",
    "PanelMinMaxScaler",
    "PanelStandardScaler",
    "ZnScoreAverager",
    "PanelPCA",
    # metrics
    "neg_mean_abs_corr",
    "panel_significance_probability",
    "sharpe_ratio",
    "sortino_ratio",
    "regression_accuracy",
    "regression_balanced_accuracy",
    "create_panel_metric",
    "correlation_coefficient",
    "regression_mcc",
    # Sequential forecasting
    "SignalOptimizer",
    "ReturnForecaster",
    # forecasting
    "NaiveRegressor",
    "LADRegressor",
    "KNNClassifier",
    "SignWeightedLADRegressor",
    "TimeWeightedLADRegressor",
    "SignWeightedLinearRegression",
    "TimeWeightedLinearRegression",
    "ModifiedLinearRegression",
    "ModifiedSignWeightedLinearRegression",
    "ModifiedTimeWeightedLinearRegression",
    "VotingRegressor",
    "VotingClassifier",
    "GlobalLocalRegression",
    # market beta estimation
    "BetaEstimator",
    # regression system
    "LADRegressionSystem",
    "RidgeRegressionSystem",
    "LinearRegressionSystem",
    "CorrelationVolatilitySystem",
    # random effects
    "RandomEffects",
    # Meta estimators
    "ProbabilityEstimator",
    "FIExtractor",
    "DataFrameTransformer",
    "CountryByCountryRegression",
]



import numbers
import numpy as np
import pandas as pd

from sklearn.linear_model import LinearRegression, Ridge, Lasso
from macrosynergy.learning.forecasting.weighted_regressors import (
    SignWeightedRegressor,
    TimeWeightedRegressor,
)


class SignWeightedLinearRegression(SignWeightedRegressor):
    def __init__(
        self,
        fit_intercept=True,
        positive=False,
        alpha=0,
        shrinkage_type="l1",
    ):
        """
        Linear regression with sign-weighted L2 loss.

        Parameters
        ----------
        fit_intercept : bool, default=True
            Whether or not to add an intercept to the model.
        positive: bool, default=False
            Whether or not to enforce non-negativity of model weights,
            with exception to a model intercept.
        alpha: float, default=0
            Shrinkage hyperparameter.
        shrinkage_type: str, default="l1"
            Type of shrinkage regularization to perform.


        Notes
        -----
        A dependent variable is modelled as a linear combination of the input features.
        The weights associated with each feature (and the intercept) are determined by
        finding the weights that minimise the weighted average squared model residuals,
        where the weighted average is based on inverse frequency of the sign of the
        dependent variable.

        By weighting the contribution of different training samples based on the
        sign of the label, the model is encouraged to learn equally from both positive and
        negative return samples, irrespective of class imbalance. If there are more
        positive targets than negative targets in the training set, then the negative
        target samples are given a higher weight in the model training process.
        The opposite is true if there are more negative targets than positive targets.
        """
        # Checks
        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")
        if not isinstance(alpha, numbers.Number) or isinstance(alpha, bool):
            raise TypeError("alpha must be a number.")
        if alpha < 0:
            raise ValueError("alpha must be non-negative.")
        if not isinstance(shrinkage_type, str):
            raise TypeError("shrinkage_type must be a string.")
        if shrinkage_type not in ["l1", "l2"]:
            raise ValueError("Invalid shrinkage_type. Must be 'l1' or 'l2'.")

        if alpha == 0:
            super().__init__(
                model=LinearRegression(fit_intercept=fit_intercept, positive=positive),
            )
        elif shrinkage_type == "l1":
            super().__init__(
                model=Lasso(
                    fit_intercept=fit_intercept, positive=positive, alpha=alpha
                ),
            )
        elif shrinkage_type == "l2":
            super().__init__(
                model=Ridge(
                    fit_intercept=fit_intercept, positive=positive, alpha=alpha
                ),
            )

        self.fit_intercept = fit_intercept
        self.positive = positive
        self.alpha = alpha
        self.shrinkage_type = shrinkage_type

    def set_params(self, **params):
        """
        Setter method to update the parameters of the `SignWeightedLinearRegression`.

        Parameters
        ----------
        **params : dict
            Dictionary of parameters to update.

        Returns
        -------
        self
            The `SignWeightedLinearRegression` instance with updated parameters.
        """
        super().set_params(**params)

        relevant_params = {"fit_intercept", "positive", "alpha", "shrinkage_type"}

        if relevant_params.intersection(params):
            if self.alpha == 0:
                self.model = LinearRegression(
                    fit_intercept=self.fit_intercept,
                    positive=self.positive,
                )
            elif self.shrinkage_type == "l1":
                self.model = Lasso(
                    fit_intercept=self.fit_intercept,
                    positive=self.positive,
                    alpha=self.alpha,
                )
            elif self.shrinkage_type == "l2":
                self.model = Ridge(
                    fit_intercept=self.fit_intercept,
                    positive=self.positive,
                    alpha=self.alpha,
                )

        return self


class TimeWeightedLinearRegression(TimeWeightedRegressor):
    def __init__(
        self,
        fit_intercept=True,
        positive=False,
        half_life=21 * 12,
        alpha=0,
        shrinkage_type="l1",
    ):
        """
        Linear regression with time-weighted L2 loss.

        Parameters
        ----------
        fit_intercept : bool, default=True
            Whether or not to add an intercept to the model.
        positive: bool, default=False
            Whether or not to enforce non-negativity of model weights,
            with exception to a model intercept.
        half_life: int, default=21 * 12
            Half-life of the exponential decay function used to weight the
            contribution of different training samples based on the time of
            the sample.
        alpha: float, default=0
            Shrinkage hyperparameter.
        shrinkage_type: str, default="l1"
            Type of shrinkage regularization to perform.

        Notes
        -----
        A dependent variable is modelled as a linear combination of the input features.
        The weights associated with each feature (and the intercept) are determined by
        finding the weights that minimise the weighted average squared model residuals,
        where the weighted average is based on an exponential decay specified by the
        half-life parameter, where more recent samples are given higher weight.

        By weighting the contribution of different training samples based on the
        timestamp, the model is encouraged to prioritise more recent samples in the
        model training process. The half-life denotes the number of time periods in units
        of the native dataset frequency for the weight attributed to the most recent sample
        (one) to decay by half.
        """
        # Checks
        if not isinstance(fit_intercept, bool):
            raise TypeError("fit_intercept must be a boolean.")
        if not isinstance(positive, bool):
            raise TypeError("positive must be a boolean.")
        if not isinstance(alpha, numbers.Number) or isinstance(alpha, bool):
            raise TypeError("alpha must be a number.")
        if alpha < 0:
            raise ValueError("alpha must be non-negative.")
        if not isinstance(shrinkage_type, str):
            raise TypeError("shrinkage_type must be a string.")
        if shrinkage_type not in ["l1", "l2"]:
            raise ValueError("Invalid shrinkage_type. Must be 'l1' or 'l2'.")
        if not isinstance(half_life, numbers.Number) or isinstance(half_life, bool):
            raise TypeError("half_life must be a number.")
        if half_life <= 0:
            raise ValueError("half_life must be positive.")

        if alpha == 0:
            model = LinearRegression(fit_intercept=fit_intercept, positive=positive)
        elif shrinkage_type == "l1":
            model = Lasso(fit_intercept=fit_intercept, positive=positive, alpha=alpha)
        elif shrinkage_type == "l2":
            model = Ridge(fit_intercept=fit_intercept, positive=positive, alpha=alpha)

        super().__init__(model=model, half_life=half_life)
        self.fit_intercept = fit_intercept
        self.positive = positive
        self.alpha = alpha
        self.shrinkage_type = shrinkage_type

    def set_params(self, **params):
        """
        Setter method to update the parameters of the `TimeWeightedLinearRegression`.

        Parameters
        ----------
        **params : dict
            Dictionary of parameters to update.

        Returns
        -------
        self
            The `TimeWeightedLinearRegression` instance with updated parameters.
        """
        super().set_params(**params)

        relevant_params = {"fit_intercept", "positive", "alpha", "shrinkage_type"}

        if relevant_params.intersection(params):
            if self.alpha == 0:
                self.model = LinearRegression(
                    fit_intercept=self.fit_intercept,
                    positive=self.positive,
                )
            elif self.shrinkage_type == "l1":
                self.model = Lasso(
                    fit_intercept=self.fit_intercept,
                    positive=self.positive,
                    alpha=self.alpha,
                )
            elif self.shrinkage_type == "l2":
                self.model = Ridge(
                    fit_intercept=self.fit_intercept,
                    positive=self.positive,
                    alpha=self.alpha,
                )

        return self


if __name__ == "__main__":
    import macrosynergy.management as msm
    from macrosynergy.management.simulate import make_qdf

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {
        "GBP": (
            pd.Timestamp(year=2009, month=1, day=1),
            pd.Timestamp(year=2012, month=6, day=30),
        ),
        "CAD": (
            pd.Timestamp(year=2015, month=1, day=1),
            pd.Timestamp(year=2100, month=1, day=1),
        ),
    }

    train = msm.categories_df(
        df=dfd, xcats=xcats, cids=cids, val="value", blacklist=black, freq="M", lag=1
    ).dropna()

    X_train = train.drop(columns=["XR"])
    y_train = train["XR"]

    # Fit SWLS - no regularization
    model = SignWeightedLinearRegression(
        fit_intercept=True, positive=False, alpha=0, shrinkage_type="l1"
    )
    model.fit(X_train, y_train)
    print(f"Intercept: {model.intercept_}, Coefficients: {model.coef_}")

    # Fit SWLS - regularization
    model = SignWeightedLinearRegression(
        fit_intercept=True, positive=False, alpha=0.01, shrinkage_type="l1"
    )
    model.fit(X_train, y_train)
    print(f"Intercept: {model.intercept_}, Coefficients: {model.coef_}")

    # Fit TWLS - no regularization
    model = TimeWeightedLinearRegression(
        fit_intercept=True, positive=False, half_life=36, alpha=0, shrinkage_type="l1"
    )
    model.fit(X_train, y_train)
    print(f"Intercept: {model.intercept_}, Coefficients: {model.coef_}")

    # Fit TWLS - regularization
    model = TimeWeightedLinearRegression(
        fit_intercept=True,
        positive=False,
        half_life=36,
        alpha=0.01,
        shrinkage_type="l1",
    )
    model.fit(X_train, y_train)
    print(f"Intercept: {model.intercept_}, Coefficients: {model.coef_}")



from typing import List, Optional, Dict

from pandas import DataFrame

from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.panel import panel_calculator


def cross_asset_effects(
        df: QuantamentalDataFrame,
        cids: List[str],
        effect_name: str,
        signal_xcats: Dict[str, str],
        weights_xcats: Dict[str, str],
        signal_signs: Optional[Dict[str, int]] = None,
) -> DataFrame:
    """
    Linear combination of a set of categories with corresponding weights and, optionally, signs.
    Corresponding assets' volatilities are generally used as weights.

    Parameters
    ----------
    df : str
        QuantamentalDataFrame with time-series of categories for both values and weights for all cross-sections.
    cids : List[str]
        List of cross-sections to compute the new quantamental category for.
    effect_name : str
        Name of the new quantamental xcat.
    signal_xcats : Dict[str, str]
        Dictionary of asset class names and related signals' time-series specified as xcats, part of df.
    weights_xcats : Dict[str, str]
        Dictionary of asset class names and related weights' time-series specified as xcats, part of df.
    signal_signs : Dict[str, int], optional
        Dictionary of asset class names and related signs in form of +1 / -1.
        Default is None, hence we assume all components contribute positively and proportionately to the final average

    Returns
    -------
    pd.DataFrame
    """
    assert isinstance(signal_xcats, dict), "Please provide a dictionary for assets' xcats"
    assert isinstance(weights_xcats, dict), "Please provide a dictionary for assets' weights"
    assert signal_signs is None or isinstance(signal_signs,
                                              dict), "Please provide a dictionary for assets' signs if needed."

    assert set(signal_xcats.keys()) == set(
        weights_xcats.keys()), "The keys of provided dictionaries for xcats and weights do not match."
    assert len(signal_xcats) == len(
        weights_xcats), "The size of provided dictionaries for xcats and weights do not match."

    if signal_signs is not None:
        assert set(signal_xcats.keys()) == set(
            signal_signs.keys()), "The keys of provided dictionaries for xcats and signs do not match."
        assert len(signal_xcats) == len(
            signal_signs), "The size of provided dictionaries for xcats and signs do not match."
    else:
        signal_signs = {
            k: 1 for k, v in signal_xcats.items()
        }

    weighted_parts_calc = [
        f"( {k.upper()}_SHARE ) * ( {str(signal_signs.get(k))} ) * {v}" for k, v in signal_xcats.items()
    ]

    calcs = [
                # Computing the total weight assigned across xcat elements
                f"WSUM = {' + '.join(weights_xcats.values())}"
            ] + [
                # Computing each category's share
                f"{k.upper()}_SHARE = {v} / WSUM" for k, v in weights_xcats.items()
            ] + [
                # Computing the final indicator as weighted average
                f"{effect_name} = {' + '.join(weighted_parts_calc)}"
            ]

    return panel_calculator(df, calcs=calcs, cids=cids)
